from django_bulk_lifecycle.enums import DEFAULT_PRIORITY


def hook(event, *, model, condition=None, priority=DEFAULT_PRIORITY):
    """
    Decorator to annotate a method with multiple lifecycle hook registrations.
    If no priority is provided, uses Priority.NORMAL (50).
    """

    def decorator(fn):
        if not hasattr(fn, "lifecycle_hooks"):
            fn.lifecycle_hooks = []
        fn.lifecycle_hooks.append((model, event, condition, priority))
        return fn

    return decorator



import logging

from django.db import transaction

from django_bulk_lifecycle.conditions import HookCondition
from django_bulk_lifecycle.registry import get_hooks, register_hook

logger = logging.getLogger(__name__)


class TriggerHandlerMeta(type):
    _registered = set()

    def __new__(mcs, name, bases, namespace):
        cls = super().__new__(mcs, name, bases, namespace)
        for method_name, method in namespace.items():
            if hasattr(method, "lifecycle_hooks"):
                for model_cls, event, condition, priority in method.lifecycle_hooks:
                    key = (model_cls, event, cls, method_name)
                    if key in TriggerHandlerMeta._registered:
                        logger.debug(
                            "Skipping duplicate registration for %s.%s on %s.%s",
                            cls.__name__,
                            method_name,
                            model_cls.__name__,
                            event,
                        )
                    else:
                        register_hook(
                            model=model_cls,
                            event=event,
                            handler_cls=cls,
                            method_name=method_name,
                            condition=condition,
                            priority=priority,
                        )
                        TriggerHandlerMeta._registered.add(key)
                        logger.debug(
                            "Registered hook %s.%s → %s.%s (cond=%r, prio=%s)",
                            model_cls.__name__,
                            event,
                            cls.__name__,
                            method_name,
                            condition,
                            priority,
                        )
        return cls


class TriggerHandler(metaclass=TriggerHandlerMeta):
    @classmethod
    def handle(
        cls,
        event: str,
        model: type,
        *,
        new_records: list = None,
        old_records: list = None,
        **kwargs,
    ) -> None:
        # Prepare hook list and log names
        hooks = get_hooks(model, event)

        # Sort hooks by priority (ascending: lower number = higher priority)
        hooks = sorted(hooks, key=lambda x: x[3])

        hook_names = [f"{h.__name__}.{m}" for h, m, _, _ in hooks]
        logger.debug(
            "Found %d hooks for %s.%s: %s",
            len(hooks),
            model.__name__,
            event,
            hook_names,
        )

        def _process():
            # Ensure new_records is a list
            new_records_local = new_records or []

            # Normalize old_records: ensure list and pad with None
            old_records_local = list(old_records) if old_records else []
            if len(old_records_local) < len(new_records_local):
                old_records_local += [None] * (
                    len(new_records_local) - len(old_records_local)
                )

            logger.debug(
                "ℹ️  bulk_lifecycle.handle() start: model=%s event=%s new_count=%d old_count=%d",
                model.__name__,
                event,
                len(new_records_local),
                len(old_records_local),
            )

            for handler_cls, method_name, condition, priority in hooks:
                logger.debug(
                    "→ evaluating hook %s.%s (cond=%r, prio=%s)",
                    handler_cls.__name__,
                    method_name,
                    condition,
                    priority,
                )

                # Evaluate condition
                passed = True
                if condition is not None:
                    if isinstance(condition, HookCondition):
                        cond_info = getattr(condition, "__dict__", str(condition))
                        logger.debug(
                            "   [cond-info] %s.%s → %r",
                            handler_cls.__name__,
                            method_name,
                            cond_info,
                        )

                        checks = []
                        for new, old in zip(new_records_local, old_records_local):
                            field_name = getattr(condition, "field", None) or getattr(
                                condition, "field_name", None
                            )
                            if field_name:
                                actual_val = getattr(new, field_name, None)
                                expected = getattr(
                                    condition, "expected_value", None
                                ) or getattr(condition, "value", None)
                                logger.debug(
                                    "   [field-lookup] %s.%s → field=%r actual=%r expected=%r",
                                    handler_cls.__name__,
                                    method_name,
                                    field_name,
                                    actual_val,
                                    expected,
                                )
                            result = condition.check(new, old)
                            checks.append(result)
                            logger.debug(
                                "   [cond-check] %s.%s → new=%r old=%r => %s",
                                handler_cls.__name__,
                                method_name,
                                new,
                                old,
                                result,
                            )
                        passed = any(checks)
                        logger.debug(
                            "   [cond-summary] %s.%s any-passed=%s",
                            handler_cls.__name__,
                            method_name,
                            passed,
                        )
                    else:
                        # Legacy callable conditions
                        passed = condition(
                            new_records=new_records_local,
                            old_records=old_records_local,
                        )
                        logger.debug(
                            "   [legacy-cond] %s.%s → full-list => %s",
                            handler_cls.__name__,
                            method_name,
                            passed,
                        )

                if not passed:
                    logger.debug(
                        "↳ skipping %s.%s (condition not met)",
                        handler_cls.__name__,
                        method_name,
                    )
                    continue

                # Instantiate & invoke handler method
                handler = handler_cls()
                method = getattr(handler, method_name)
                logger.info(
                    "✨ invoking %s.%s on %d record(s)",
                    handler_cls.__name__,
                    method_name,
                    len(new_records_local),
                )
                try:
                    method(
                        new_records=new_records_local,
                        old_records=old_records_local,
                        **kwargs,
                    )
                except Exception:
                    logger.exception(
                        "❌ exception in %s.%s",
                        handler_cls.__name__,
                        method_name,
                    )

            logger.debug(
                "✔️  bulk_lifecycle.handle() complete for %s.%s",
                model.__name__,
                event,
            )

        # Defer if in atomic block and event is after_*
        conn = transaction.get_connection()
        if conn.in_atomic_block and event.startswith("after_"):
            logger.debug(
                "Deferring hook execution until after transaction commit for event '%s'",
                event,
            )
            transaction.on_commit(_process)
        else:
            _process()


import logging

logger = logging.getLogger(__name__)


class HookCondition:
    def check(self, instance, original_instance=None):
        raise NotImplementedError

    def __call__(self, instance, original_instance=None):
        return self.check(instance, original_instance)

    def __and__(self, other):
        return AndCondition(self, other)

    def __or__(self, other):
        return OrCondition(self, other)

    def __invert__(self):
        return NotCondition(self)


class WhenFieldValueIsNot(HookCondition):
    def __init__(self, field, unexpected_value, only_on_change=False):
        self.field = field
        self.unexpected_value = unexpected_value
        self.only_on_change = only_on_change

    def check(self, instance, original_instance=None):
        current = getattr(instance, self.field)
        logger.debug("%s current=%r, original=%r",
                     self.field,
                     current,
                     getattr(original_instance, self.field, None))
        if self.only_on_change:
            if original_instance is None:
                return False
            previous = getattr(original_instance, self.field)
            return previous == self.unexpected_value and current != self.unexpected_value
        else:
            return current != self.unexpected_value


class WhenFieldValueIs(HookCondition):
    def __init__(self, field, expected_value, only_on_change=False):
        self.field = field
        self.expected_value = expected_value
        self.only_on_change = only_on_change

    def check(self, instance, original_instance=None):
        current = getattr(instance, self.field)
        logger.debug("%s current=%r, original=%r",
                     self.field,
                     current,
                     getattr(original_instance, self.field, None))
        if self.only_on_change:
            if original_instance is None:
                return False
            previous = getattr(original_instance, self.field)
            return previous != self.expected_value and current == self.expected_value
        else:
            return current == self.expected_value


class WhenFieldHasChanged(HookCondition):
    def __init__(self, field, has_changed=True):
        self.field = field
        self.has_changed = has_changed

    def check(self, instance, original_instance=None):
        if not original_instance:
            return False
        return (getattr(instance, self.field) != getattr(original_instance, self.field)) == self.has_changed


class WhenFieldValueWas(HookCondition):
    def __init__(self, field, expected_value, only_on_change=False):
        """
        Check if a field's original value was `expected_value`.
        If only_on_change is True, only return True when the field has changed away from that value.
        """
        self.field = field
        self.expected_value = expected_value
        self.only_on_change = only_on_change

    def check(self, instance, original_instance=None):
        if original_instance is None:
            return False
        previous = getattr(original_instance, self.field)
        if self.only_on_change:
            current = getattr(instance, self.field)
            return previous == self.expected_value and current != self.expected_value
        else:
            return previous == self.expected_value


class WhenFieldValueChangesTo(HookCondition):
    def __init__(self, field, expected_value):
        """
        Check if a field's value has changed to `expected_value`.
        Only returns True when original value != expected_value and current value == expected_value.
        """
        self.field = field
        self.expected_value = expected_value

    def check(self, instance, original_instance=None):
        if original_instance is None:
            return False
        previous = getattr(original_instance, self.field)
        current = getattr(instance, self.field)
        return previous != self.expected_value and current == self.expected_value


class AndCondition(HookCondition):
    def __init__(self, cond1, cond2):
        self.cond1 = cond1
        self.cond2 = cond2

    def check(self, instance, original_instance=None):
        return self.cond1.check(instance, original_instance) and self.cond2.check(instance, original_instance)


class OrCondition(HookCondition):
    def __init__(self, cond1, cond2):
        self.cond1 = cond1
        self.cond2 = cond2

    def check(self, instance, original_instance=None):
        return self.cond1.check(instance, original_instance) or self.cond2.check(instance, original_instance)


class NotCondition(HookCondition):
    def __init__(self, cond):
        self.cond = cond

    def check(self, instance, original_instance=None):
        return not self.cond.check(instance, original_instance)



from django_bulk_lifecycle.manager import BulkLifecycleManager
from django.db import models, transaction


class LifecycleModelMixin(models.Model):
    objects = BulkLifecycleManager()

    class Meta:
        abstract = True

    def delete(self, *args, **kwargs):
        self.before_delete()

        with transaction.atomic():
            result = super().delete(*args, **kwargs)

        self.after_delete()

        return result

    def before_delete(self):
        pass

    def after_delete(self):
        pass



from enum import IntEnum


class Priority(IntEnum):
    """
    Named priorities for django-bulk-lifecycle hooks.

    Lower values run earlier (higher priority).
    Hooks are sorted in ascending order.
    """

    HIGHEST = 0  # runs first
    HIGH = 25  # runs early
    NORMAL = 50  # default ordering
    LOW = 75  # runs later
    LOWEST = 100  # runs last
